# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from unittest import TestCase

from nose.tools import istest

from nisteag.entropy import AnagramError, DictionaryError
from nisteag.token.requirements.memorized import (
    Level1Checker,
    Level2Checker,
    NullThrottler,
    WeakTokenError,
)


class FailingThrottler(object):
    def check(self, username, token):
        raise RuntimeError('too many attempts')


class Level1CheckerTest(TestCase):
    def setUp(self):
        self.checker = Level1Checker(NullThrottler())

    @istest
    def checks_a_strong_token(self):
        self.checker.check('This Is a BIG and relev4nt passwurd!!!')

    @istest
    def fails_a_too_weak_token(self):
        with self.assertRaises(WeakTokenError):
            self.checker.check('12345')

    @istest
    def checks_a_strong_unicode_token(self):
        self.checker.check('á1234')

    @istest
    def fails_an_existing_token(self):
        token = 'Some Exist1ng Tok3n!!!'

        with self.assertRaises(DictionaryError):
            self.checker.check(token, dictionary=['something', token])

    @istest
    def fails_an_anagram_token(self):
        token = 'This Is a BIG and relev4nt passwurd!!!'

        with self.assertRaises(AnagramError):
            self.checker.check(token, username=reversed(token))

    @istest
    def fails_if_throttler_fails(self):
        checker = Level1Checker(FailingThrottler())

        with self.assertRaises(RuntimeError):
            checker.check('kjhdgfguier89324!!!')


class Level2CheckerTest(TestCase):
    def setUp(self):
        self.checker = Level2Checker(NullThrottler())

    @istest
    def checks_a_strong_token(self):
        self.checker.check('This Is a BIG and relev4nt passwurd!!!')

    @istest
    def fails_a_too_weak_token(self):
        with self.assertRaises(WeakTokenError):
            self.checker.check('1234567')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import string
from unittest import TestCase

from nose.tools import istest

from nisteag.entropy import (
    AnagramError,
    DictionaryError,
    EmptyTokenError,
    EntropyCalculator,
)


class EntropyCalculatorTest(TestCase):
    def setUp(self):
        self.calculator = EntropyCalculator()
        self.dictionary = [
            'some pass',
            'another pass',
        ]

    @istest
    def calculates_for_empty_token(self):
        result = self.calculator.calculate('')

        self.assertEqual(result, 0)

    @istest
    def calculates_for_single_ascii_character(self):
        result = self.calculator.calculate('a')

        self.assertEqual(result, 4)

    @istest
    def calculates_for_single_ascii_upper_character(self):
        result = self.calculator.calculate('A')

        self.assertEqual(result, 4)

    @istest
    def calculates_for_2_ascii_characters(self):
        result = self.calculator.calculate(string.printable[:2])

        self.assertEqual(result, 6)

    @istest
    def calculates_for_3_ascii_characters(self):
        result = self.calculator.calculate(string.printable[:3])

        self.assertEqual(result, 8)

    @istest
    def calculates_for_4_ascii_characters(self):
        result = self.calculator.calculate(string.printable[:4])

        self.assertEqual(result, 10)

    @istest
    def calculates_for_6_ascii_characters(self):
        result = self.calculator.calculate(string.printable[:6])

        self.assertEqual(result, 14)

    @istest
    def calculates_for_8_ascii_characters(self):
        result = self.calculator.calculate(string.printable[:8])

        self.assertEqual(result, 18)

    @istest
    def calculates_for_9_ascii_characters(self):
        result = self.calculator.calculate(string.printable[:9])

        self.assertEqual(result, 19.5)

    @istest
    def calculates_for_21_ascii_characters(self):
        token = (string.ascii_letters * 2)[:21]
        result = self.calculator.calculate(token)

        self.assertEqual(result, 37)

    @istest
    def calculates_for_single_unicode_character(self):
        result = self.calculator.calculate('á')

        self.assertEqual(result, 12.26)

    @istest
    def calculates_for_single_ascii_byte(self):
        result = self.calculator.calculate('a'.encode('ascii'))

        self.assertEqual(result, 4)

    @istest
    def calculates_for_single_utf8_byte(self):
        result = self.calculator.calculate('á'.encode('utf-8'))

        self.assertEqual(result, 12.26)

    @istest
    def calculates_for_single_latin1_byte(self):
        result = self.calculator.calculate('á'.encode('latin1'))

        self.assertEqual(result, 12.26)

    @istest
    def calculates_for_mixed_case(self):
        result = self.calculator.calculate('abcA')

        self.assertEqual(result, 12)

    @istest
    def calculates_for_small_mixed_case(self):
        result = self.calculator.calculate('abC')

        self.assertEqual(result, 8)

    @istest
    def calculates_for_lower_and_digit(self):
        result = self.calculator.calculate('abc1')

        self.assertEqual(result, 12)

    @istest
    def calculates_for_upper_and_digit(self):
        result = self.calculator.calculate('ABC1')

        self.assertEqual(result, 12)

    @istest
    def calculates_for_upper_and_special(self):
        result = self.calculator.calculate('ABC*')

        self.assertEqual(result, 12)

    @istest
    def calculates_for_lower_and_upper_and_special_and_number(self):
        result = self.calculator.calculate('aA*1')

        self.assertEqual(result, 12)

    @istest
    def calculates_for_mixed_case_5(self):
        result = self.calculator.calculate('Abcde')

        self.assertEqual(result, 15)

    @istest
    def calculates_for_mixed_case_9(self):
        result = self.calculator.calculate('Abcdefghi')

        self.assertEqual(result, 25.5)

    @istest
    def calculates_for_3_ascii_characters_with_dictionary(self):
        result = self.calculator.calculate(
            string.printable[:3], dictionary=self.dictionary)

        self.assertEqual(result, 8)

    @istest
    def fails_to_calculate_if_token_is_in_dictionary(self):
        with self.assertRaises(DictionaryError):
            self.calculator.calculate(
                self.dictionary[1], dictionary=self.dictionary)

    @istest
    def calculates_for_4_ascii_characters_with_dictionary(self):
        result = self.calculator.calculate(
            string.printable[:4], self.dictionary)

        self.assertEqual(result, 14)

    @istest
    def calculates_for_5_ascii_characters_with_dictionary(self):
        result = self.calculator.calculate(
            string.printable[:5], self.dictionary)

        self.assertEqual(result, 17)

    @istest
    def calculates_for_7_ascii_characters_with_dictionary(self):
        result = self.calculator.calculate(
            string.printable[:7], self.dictionary)

        self.assertEqual(result, 22)

    @istest
    def calculates_for_9_ascii_characters_with_dictionary(self):
        result = self.calculator.calculate(
            string.printable[:9], self.dictionary)

        self.assertEqual(result, 25)

    @istest
    def calculates_for_10_ascii_characters_with_dictionary(self):
        result = self.calculator.calculate(
            string.printable[:10], self.dictionary)

        self.assertEqual(result, 26)

    @istest
    def calculates_for_21_ascii_characters_with_dictionary(self):
        token = (string.ascii_letters * 2)[:21]
        result = self.calculator.calculate(token, self.dictionary)

        self.assertEqual(result, 37)

    @istest
    def fails_to_calculate_if_token_is_anagram_of_username(self):
        token = 'silent'
        username = 'listen'

        with self.assertRaises(AnagramError):
            self.calculator.calculate(token, username=username)

    @istest
    def calculates_for_3_ascii_characters_with_username(self):
        result = self.calculator.calculate(
            string.printable[:3], username='joe')

        self.assertEqual(result, 8)

    @istest
    def calculates_for_3_ascii_characters_with_dict_and_username(self):
        result = self.calculator.calculate(
            string.printable[:3], self.dictionary, 'joe')

        self.assertEqual(result, 10)



"""Requirements for Memorized Secret Tokens.

See section 6.3.1.1. Single Token Authentication, table 6, page 51, of the NIST
document.

"""

from abc import ABCMeta, abstractmethod

from nisteag.entropy import EntropyCalculator, EntropyError


class WeakTokenError(EntropyError):
    """Raised when the token is too weak."""


class BaseThrottler(object):
    """Base class for implementation of throttling mechanisms.

    The only method that needs to be implemented is
    :meth:`BaseThrottler.check`.

    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def check(self, username, token):
        raise NotImplementedError()  # pragma: no cover


class NullThrottler(BaseThrottler):
    """A null throttler, i.e., it does not actually do anything.

    The check from this class always passes, it's used for experimentation or
    testing purposes only. Use it at your own risk.

    """

    def check(self, username, token):
        pass


class BaseMemorizedChecker(object):
    MINIMUM_ENTROPY = 0

    def __init__(self, throttler):
        """Initializes the instance.

        :param BaseThrottler throttler: The throttler that will be used for
            making sure that failed attempts are limited.

        """
        self.calculator = EntropyCalculator()
        self.throttler = throttler

    def check(self, token, dictionary=None, username=None):
        """Checks a token to make sure it meets the requirements.

        :param str token: The token for the calculation. May be a password, a
            pass-phrase and the like.
        :param sequence dictionary: An optional dictionary as a sequence,
            against which the token will be tested, if provided. If you opt by
            using it, provide a dictionary of at least 50,000 items.
            Default: `None`.
        :param str username: An optional username to be used for anagram
            checking. Default: `None`.

        """
        self.throttler.check(username, token)

        bits = self.calculator.calculate(token, dictionary, username)

        if bits < self.MINIMUM_ENTROPY:
            raise WeakTokenError('Token is too weak.')


class Level1Checker(BaseMemorizedChecker):
    MINIMUM_ENTROPY = 14


class Level2Checker(BaseMemorizedChecker):
    MINIMUM_ENTROPY = 18



from unittest import TestCase

from mock import patch
from nose.tools import istest

from nisteag.scripts.checkers import check_entropy


class CheckEntropyTest(TestCase):
    @istest
    @patch('nisteag.scripts.checkers.print')
    @patch('nisteag.scripts.checkers.sys')
    def checks_entropy(self, mock_sys, mock_print):
        mock_sys.argv = ['check-entropy', 'abcd']

        check_entropy()

        mock_print.assert_called_once_with(10.0)

    @istest
    @patch('nisteag.scripts.checkers.print')
    @patch('nisteag.scripts.checkers.sys')
    @patch('nisteag.scripts.checkers.getpass')
    def checks_entropy_with_safe_pass(self, mock_pass, mock_sys, mock_print):
        mock_sys.argv = ['check-entropy']
        mock_pass.return_value = 'abcd'

        check_entropy()

        mock_print.assert_called_once_with(10.0)



from setuptools import setup, find_packages
import sys, os

version = '0.2.0'

setup(
    name='nisteag',
    version=version,
    description="An implementation of recommendations from the NIST Electronic Authentication Guideline",
    long_description="""\
""",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Security',
    ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords='nist electronic authentication guideline password security',
    author='Diogo Baeder',
    author_email='diogo.baeder@yougov.com',
    url='https://github.com/yougov/nisteag',
    license='MIT',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests', 'docs']),
    package_dir={'nisteag': 'nisteag'},
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'chardet>=2.3.0',
        'six>=1.10.0',
    ],
    entry_points={
        'console_scripts': [
            'check-entropy = nisteag.scripts.checkers:check_entropy'
        ],
    },
)



from __future__ import print_function

import sys
from getpass import getpass

from nisteag.entropy import EntropyCalculator


def check_entropy():
    calculator = EntropyCalculator()

    try:
        token = sys.argv[1]
    except IndexError:
        token = getpass()

    print(calculator.calculate(token))



"""This module is related to token entropy calculation and the like.

See Appendix A of the NIST document (page 103) if you want to understand how
the calculation is done.

"""

import math
import string

import chardet
import six


KEYBOARD_SIZE = 94
UNICODE_SIZE = 1114111


ALL_SET = set(string.printable)
LOWER_SET = set(string.ascii_lowercase)
UPPER_SET = set(string.ascii_uppercase)
DIGIT_SET = set(string.digits)
SPECIAL_SET = ALL_SET - (LOWER_SET | UPPER_SET | DIGIT_SET)
SETS = sorted((LOWER_SET, UPPER_SET, DIGIT_SET, SPECIAL_SET))


MAX_COMPOSITION_BITS = 6
MAX_DICTIONARY_BITS = 6
MIN_DICT_USER_BITS = 10


class EntropyError(Exception):
    """Base exception class for entropy errors."""


class EmptyTokenError(EntropyError):
    """Raised when the token is empty."""


class DictionaryError(EntropyError):
    """Raised when there's an error regarding the terms dictionary provided."""


class AnagramError(EntropyError):
    """Raised when the token is an anagram of a username."""


class EntropyCalculator(object):
    """This class is used to calculate entropies for tokens."""

    def _get_alphabet_size(self, token):
        try:
            token.encode('ascii')
        except UnicodeEncodeError:
            return UNICODE_SIZE
        return KEYBOARD_SIZE

    def calculate(self, token, dictionary=None, username=None):
        """Calculates the entropy for a given token.

        :param str token: The token for the calculation. May be a password, a
            pass-phrase and the like.
        :param sequence dictionary: An optional dictionary as a sequence,
            against which the token will be tested, if provided. If you opt by
            using it, provide a dictionary of at least 50,000 items.
            Default: `None`.
        :param str username: An optional username to be used for anagram
            checking. Default: `None`.

        """

        if not token:
            return 0

        if dictionary and token in dictionary:
            raise DictionaryError('Token exists in the dictionary provided.')

        if username and sorted(username) == sorted(token):
            raise AnagramError('Token must not be an anagram of username.')

        if not isinstance(token, six.text_type):
            token = self._decode_token(token)

        bits = self._calculate_bits(token)

        size = self._get_alphabet_size(token)
        coef = math.log(size, 2) / math.log(KEYBOARD_SIZE, 2)

        bits += self._get_composition_additional(token)
        bits += self._get_dictionary_additional(token, dictionary)

        if username and dictionary:
            bits = max(bits, MIN_DICT_USER_BITS)

        return round(bits * coef, 2)

    def _get_dictionary_additional(self, token, dictionary):
        length = len(token)
        if not dictionary or length < 4 or length >= 20:
            return 0
        if length > 8:
            return MAX_DICTIONARY_BITS - (length - 8) / 2.0
        return min(length, MAX_DICTIONARY_BITS)

    def _get_composition_additional(self, token):
        token_set = set(token)
        token_length = len(token)
        composition = 0

        if self._is_composed(token_set) and token_length > 3:
            composition = min(token_length - 2, MAX_COMPOSITION_BITS)

        return composition

    def _is_composed(self, token_set):
        for i, first_set in enumerate(SETS):
            for j, second_set in enumerate(SETS):
                composes = (
                    i != j
                    and token_set & first_set
                    and token_set & second_set
                )
                if composes:
                    return True
        return False

    def _calculate_bits(self, token):
        bits = 0

        for i, char in enumerate(token):
            if i == 0:
                char_bits = 4
            elif i < 8:
                char_bits = 2
            elif i < 20:
                char_bits = 1.5
            else:
                char_bits = 1
            bits += char_bits

        return bits

    def _decode_token(self, token):
        try:
            return token.decode('utf-8')
        except UnicodeDecodeError:
            charset = chardet.detect(token)
            return token.decode(charset['encoding'])
